@@ -18,18 +18,11 @@
 
 import unittest
 from unittest import mock
-import math
 import threading
 from contextlib import contextmanager
-from functools import wraps
-import gc
 import os.path
 import os
-import sys
 import time
-import io
-import platform
-import ctypes
 from concurrent.futures import Future
 
 os.environ["PATH"] = os.path.dirname(__file__) + os.pathsep + os.environ["PATH"]
@@ -49,6 +42,8 @@ else:
 TESTVID = os.path.join(os.path.dirname(__file__), 'test.webm')
 TESTSRT = os.path.join(os.path.dirname(__file__), 'sub_test.srt')
 MPV_ERRORS = [ l(ec) for ec, l in mpv.ErrorCode.EXCEPTION_DICT.items() if l ]
+SKIP_TESTS = os.environ.get('PY_MPV_SKIP_TESTS', '').split()
+
 
 def timed_print():
     start_time = time.time()
@@ -56,6 +51,7 @@ def timed_print():
         td = time.time() - start_time
         print('{:.3f} [{}] {}: {}'.format(td, level, prefix, text), flush=True)
 
+
 class MpvTestCase(unittest.TestCase):
     def setUp(self):
         self.disp = Xvfb()
@@ -66,6 +62,7 @@ class MpvTestCase(unittest.TestCase):
         self.m.terminate()
         self.disp.stop()
 
+
 class TestProperties(MpvTestCase):
     @contextmanager
     def swallow_mpv_errors(self, exception_exceptions=[]):
@@ -295,6 +292,7 @@ class ObservePropertyTest(MpvTestCase):
             mock.call('slang', ['ru'])],
             any_order=True)
 
+
 class KeyBindingTest(MpvTestCase):
     def test_register_direct_cmd(self):
         self.m.register_key_binding('a', 'playlist-clear')
@@ -441,6 +439,7 @@ class KeyBindingTest(MpvTestCase):
         handler1.assert_has_calls([])
         handler2.assert_has_calls([ mock.call() ])
 
+
 class TestStreams(unittest.TestCase):
     def test_python_stream(self):
         handler = mock.Mock()
@@ -536,6 +535,116 @@ class TestStreams(unittest.TestCase):
         m.terminate()
         disp.stop()
 
+    def test_stream_open_exception(self):
+        disp = Xvfb()
+        disp.start()
+        m = mpv.MPV(vo=testvo, video=False)
+
+        @m.register_stream_protocol('raiseerror')
+        def open_fn(uri):
+            raise SystemError()
+
+        waiting = threading.Semaphore()
+        result = Future()
+        def run():
+            result.set_running_or_notify_cancel()
+            try:
+                waiting.release()
+                m.wait_for_playback()
+                result.set_result(False)
+            except SystemError:
+                result.set_result(True)
+            except Exception:
+                result.set_result(False)
+
+        t = threading.Thread(target=run, daemon=True)
+        t.start()
+
+        with waiting:
+            time.sleep(0.2)
+            m.play('raiseerror://foo')
+
+        m.wait_for_playback(catch_errors=False)
+        try:
+            assert result.result()
+        finally:
+            m.terminate()
+            disp.stop()
+
+    def test_python_stream_exception(self):
+        disp = Xvfb()
+        disp.start()
+        m = mpv.MPV(vo=testvo)
+
+        @m.python_stream('foo')
+        def foo_gen():
+            with open(TESTVID, 'rb') as f:
+                yield f.read(100)
+                raise SystemError()
+
+        waiting = threading.Semaphore()
+        result = Future()
+        def run():
+            result.set_running_or_notify_cancel()
+            try:
+                waiting.release()
+                m.wait_for_playback()
+                result.set_result(False)
+            except SystemError:
+                result.set_result(True)
+            except Exception:
+                result.set_result(False)
+
+        t = threading.Thread(target=run, daemon=True)
+        t.start()
+
+        with waiting:
+            time.sleep(0.2)
+            m.play('python://foo')
+
+        m.wait_for_playback(catch_errors=False)
+        try:
+            assert result.result()
+        finally:
+            m.terminate()
+            disp.stop()
+
+    def test_stream_open_forward(self):
+        disp = Xvfb()
+        disp.start()
+        m = mpv.MPV(vo=testvo, video=False)
+
+        @m.register_stream_protocol('raiseerror')
+        def open_fn(uri):
+            raise ValueError()
+
+        waiting = threading.Semaphore()
+        result = Future()
+        def run():
+            result.set_running_or_notify_cancel()
+            try:
+                waiting.release()
+                m.wait_for_playback()
+                result.set_result(True)
+            except Exception:
+                result.set_result(False)
+
+        t = threading.Thread(target=run, daemon=True)
+        t.start()
+
+        with waiting:
+            time.sleep(0.2)
+            m.play('raiseerror://foo')
+
+        m.wait_for_playback(catch_errors=False)
+        try:
+            assert result.result()
+        finally:
+            m.terminate()
+            disp.stop()
+
+
+
 class TestLifecycle(unittest.TestCase):
     def test_create_destroy(self):
         thread_names = lambda: [ t.name for t in threading.enumerate() ]
@@ -577,8 +686,8 @@ class TestLifecycle(unittest.TestCase):
                 mock.call({'event': 'start-file', 'playlist_entry_id': 1}),
                 mock.call({'event': 'end-file', 'reason': 'error', 'playlist_entry_id': 1, 'file_error': 'no audio or video data played'})
             ], any_order=True)
+        time.sleep(1)
         handler.reset_mock()
-
         m.terminate()
         handler.assert_not_called()
 
@@ -600,6 +709,7 @@ class TestLifecycle(unittest.TestCase):
         t.start()
         time.sleep(1)
         m.terminate()
+        time.sleep(1)
         t.join()
         self.disp.stop()
         assert result.result()
@@ -620,13 +730,13 @@ class TestLifecycle(unittest.TestCase):
         m.mute = True
         t.join()
         m.terminate()
+        time.sleep(1)
         handler.assert_called()
         self.disp.stop()
 
     def test_wait_for_event(self):
         self.disp = Xvfb()
         self.disp.start()
-        handler = mock.Mock()
         m = mpv.MPV(vo=testvo)
         m.play(TESTVID)
         result = Future()
@@ -649,7 +759,6 @@ class TestLifecycle(unittest.TestCase):
     def test_wait_for_property_shutdown(self):
         self.disp = Xvfb()
         self.disp.start()
-        handler = mock.Mock()
         m = mpv.MPV(vo=testvo)
         m.play(TESTVID)
         with self.assertRaises(mpv.ShutdownError):
@@ -657,12 +766,13 @@ class TestLifecycle(unittest.TestCase):
             # handle
             with m.prepare_and_wait_for_property('mute', level_sensitive=False):
                 m.terminate()
+        time.sleep(1)
         self.disp.stop()
 
-    def test_wait_for_prooperty_event_overflow(self):
+    @unittest.skipIf('test_wait_for_property_event_overflow' in SKIP_TESTS, reason="kills X-Server first")
+    def test_wait_for_property_event_overflow(self):
         self.disp = Xvfb()
         self.disp.start()
-        handler = mock.Mock()
         m = mpv.MPV(vo=testvo)
         m.play(TESTVID)
         with self.assertRaises(mpv.EventOverflowError):
@@ -677,10 +787,10 @@ class TestLifecycle(unittest.TestCase):
                         m.command_async('script-message', 'foo', 'bar')
                     except:
                         pass
+        m.terminate()
+        time.sleep(1)
         self.disp.stop()
 
-
-
     def test_wait_for_event_shutdown(self):
         self.disp = Xvfb()
         self.disp.start()
@@ -694,7 +804,6 @@ class TestLifecycle(unittest.TestCase):
     def test_wait_for_shutdown(self):
         self.disp = Xvfb()
         self.disp.start()
-        handler = mock.Mock()
         m = mpv.MPV(vo=testvo)
         m.play(TESTVID)
         with self.assertRaises(mpv.ShutdownError):
@@ -766,7 +875,6 @@ class CommandTests(MpvTestCase):
         callback.assert_any_call(None, None)
 
 
-
 class RegressionTests(MpvTestCase):
 
     def test_unobserve_property_runtime_error(self):
@@ -820,7 +928,3 @@ class RegressionTests(MpvTestCase):
         m.slang = 'ru'
         m.terminate() # needed for synchronization of event thread
         handler.assert_has_calls([mock.call('slang', ['jp']), mock.call('slang', ['ru'])])
-
-
-if __name__ == '__main__':
-    unittest.main()
